#include "cvimageviewer.h"
#include <QPicture>
#include <QDebug>
#include <QMouseEvent>
#include <QFileDialog>
#include <opencv2/imgcodecs.hpp>
#include <statisticsdialog.h>

CvImageViewer::CvImageViewer(QWidget *parent, size_t lastId) :
	QWidget(parent),
	lastId_(lastId),
	saveAction_("Save Image", nullptr),
	zoomAction_("Zoom to selection", nullptr),
	resetAction_("Reset Zoom", nullptr),
	statisticsAction_("Get selection properties", nullptr),
	roi_(0,0,0,0)
{
	qimage_.load(":/images/noimage.png");

	connect(&saveAction_, &QAction::triggered, this, &CvImageViewer::saveImage);
	connect(&zoomAction_, &QAction::triggered, this, &CvImageViewer::zoomToSelection);
	connect(&resetAction_, &QAction::triggered, this, &CvImageViewer::resetZoom);
	connect(&statisticsAction_, &QAction::triggered, this, &CvImageViewer::showSatDiag);

	imageContextMenu_.addAction(&saveAction_);
	imageContextMenu_.addAction(&zoomAction_);
	imageContextMenu_.addAction(&resetAction_);
	imageContextMenu_.addAction(&statisticsAction_);
}

CvImageViewer::~CvImageViewer()
{
}

void CvImageViewer::saveImage()
{
	QString fileName;
	if(origImage_.type() == CV_8UC3 || origImage_.type() == CV_8SC3 || origImage_.type() == CV_8UC1 || origImage_.type() == CV_8SC1)
	{
		fileName = QFileDialog::getSaveFileName(this, "Save Image", "./", "*.mat *.png" );
	}
	else
	{
		fileName = QFileDialog::getSaveFileName(this, "Save Image", "./", "*.mat" );
	}
	if(!fileName.isEmpty())
	{
		QStringList tokens = fileName.split('.');
		if(tokens.back() != "mat" && tokens.back() != "png")
			fileName.append(".mat");
		tokens = fileName.split('.');
		if(tokens.back() == "png")
		{
			imwrite(fileName.toStdString(), origImage_);
		}
		else
		{
			cv::FileStorage matf(fileName.toStdString(), cv::FileStorage::WRITE);
			matf<<"image"<<origImage_;
			matf.release();
		}
	}
}

void CvImageViewer::showSatDiag()
{
	if(origImage_.channels() > 1 || selectionRect_.width() < 2 || selectionRect_.height() < 2)
		return;

	int xA;
	int yA;
	int xB;
	int yB;
	transfromToSourceCoordinates(selectionRect_.x(), selectionRect_.y(), xA, yA);
	transfromToSourceCoordinates(selectionRect_.x()+selectionRect_.width(), selectionRect_.y()+selectionRect_.height(), xB, yB);
	selectionRect_ = QRect(0, 0, 0, 0);
	cv::Rect roi(xA, yA, xB-xA, yB-yA);

	cv::Mat roiImage = origImage_(roi);
	if(roiImage.type() != CV_64FC1)
		roiImage.convertTo(roiImage, CV_64FC1);

	std::vector<double> data;
	data.reserve(roiImage.total());
	for (cv::MatIterator_<double> it = roiImage.begin<double>(); it != roiImage.end<double>(); ++it)
		data.push_back(*it);

	StatisticsDialog diag(data, this);
	diag.show();
	diag.exec();
}

void CvImageViewer::convertImage(cv::Mat image)
{
	image_ = image;
	if(image_.type() == CV_8UC3 || image_.type() == CV_8SC3)
	{
		qimage_ = QImage(image_.data, image_.cols, image_.rows, image_.step, QImage::Format_RGB888);
		statisticsAction_.setDisabled(true);
	}
	else if(image_.type() == CV_8UC1 || image_.type() == CV_8SC1)
	{
		qimage_ = QImage(image_.data, image_.cols, image_.rows, image_.step, QImage::Format_Grayscale8);
		statisticsAction_.setDisabled(false);
	}
	else if(image_.type() == CV_32FC1 || image_.type() == CV_64FC1)
	{
		double min, max;
		cv::minMaxIdx(image_, &min, &max);
		double a = 255.0/(max - min);
		double b = 0-(min*a);
		qDebug()<<min<<max<<a<<b;
		image_.convertTo(image_, CV_8UC1, a, b);
		qimage_ = QImage(image_.data, image_.cols, image_.rows, image_.step, QImage::Format_Grayscale8);
		statisticsAction_.setDisabled(false);
	}
	else if(image_.type() == CV_32FC3 || image_.type() == CV_64FC3)
	{
		double min, max;
		cv::minMaxIdx(image_, &min, &max);
		double a = 255.0/(max - min);
		double b = 0-(min*a);
		qDebug()<<min<<max<<a<<b;
		image_.convertTo(image_, CV_8UC3, a, b);
		qimage_ = QImage(image_.data, image_.cols, image_.rows, image_.step, QImage::Format_Grayscale8);
		statisticsAction_.setDisabled(true);
	}
	else
	{
		image_.convertTo(image_, CV_8UC1, 255, 0);
		qimage_ = QImage(image_.data, image_.cols, image_.rows, image_.step, QImage::Format_Grayscale8);
		statisticsAction_.setDisabled(true);
	}
}

void CvImageViewer::setImage(Camera::Image img)
{
	origImage_=img.mat;
	qDebug()<<"viwer got"<<image_.rows<<'x'<<image_.cols<<" type "<<image_.type()<<"image from camera"<<img.cameraId;
	convertImage(img.mat);
	update();
	roi_ = cv::Rect(0, 0, image_.size().width, image_.size().height);
}

void CvImageViewer::transfromToSourceCoordinates(int inX, int inY, int& outX, int& outY)
{
	qDebug()<<roi_.x<<inX<<imgrect_.x()<<imgrect_.width()<<roi_.width;
	outX = roi_.x+(inX-imgrect_.x())/static_cast<double>(imgrect_.width())*roi_.width;
	outY = roi_.y+(inY-imgrect_.y())/static_cast<double>(imgrect_.height())*roi_.height;
}


void CvImageViewer::resetZoom()
{
	selectionRect_ = QRect(0, 0, 0, 0);
	convertImage(origImage_);
	update();
	roi_ = cv::Rect(0, 0, image_.size().width, image_.size().height);
}

void CvImageViewer::zoomToSelection()
{
	if(selectionRect_.width() > 2 && selectionRect_.height() > 2)
	{
		convertImage(origImage_);
		int xA;
		int yA;
		int xB;
		int yB;
		transfromToSourceCoordinates(selectionRect_.x(), selectionRect_.y(), xA, yA);
		transfromToSourceCoordinates(selectionRect_.x()+selectionRect_.width(), selectionRect_.y()+selectionRect_.height(), xB, yB);
		selectionRect_ = QRect(0, 0, 0, 0);
		roi_ = cv::Rect(xA, yA, xB-xA, yB-yA);
		cv::Mat cropped;
		cropped = image_(roi_);
		convertImage(cropped);
		update();
	}
}

void CvImageViewer::mousePressEvent(QMouseEvent *event)
{
	if(origImage_.data)
	{
		if(event->button() == Qt::RightButton)
		{
			saveAction_.setEnabled(origImage_.data);
			imageContextMenu_.popup(event->globalPos());
		}
		else if(origImage_.data && event->x() > imgrect_.x() && event->y() > imgrect_.y() && event->x() < imgrect_.x()+imgrect_.width() && event->y() < imgrect_.y()+imgrect_.height())
		{
			selectionStarted_ = true;
			selectionRect_.setTopLeft(event->pos());
			selectionRect_.setBottomRight(event->pos());
			int x;
			int y;
			transfromToSourceCoordinates(event->x(), event->y(), x, y);
			if(origImage_.type() == CV_32FC1)
			{
				if(x >= 0 && y >= 0 && x <= origImage_.cols && y < origImage_.rows)
					sigValue(x, y, origImage_.at<float>(y,x));
			}
			else
			{
				sigValue(x, y, 0);
			}
		}
	}
	QWidget::mousePressEvent(event);
}

void CvImageViewer::mouseMoveEvent(QMouseEvent *event)
{
	if(selectionStarted_)
	{
		if(event->x() > imgrect_.x() && event->y() > imgrect_.y() && event->x() < imgrect_.x()+imgrect_.width() && event->y() < imgrect_.y()+imgrect_.height())
		{
			selectionRect_.setBottomRight(event->pos());
			repaint();
		}
	}
}

void CvImageViewer::mouseReleaseEvent(QMouseEvent *event)
{
	selectionStarted_ = false;
	QWidget::mouseReleaseEvent(event);
}

void CvImageViewer::resizeEvent(QResizeEvent *event)
{
	selectionRect_ = QRect(0, 0, 0, 0);
	QWidget::resizeEvent(event);
}

void CvImageViewer::paintEvent(QPaintEvent* event)
{
	Q_UNUSED(event)
	QPainter painter(this);

	double ratio = qimage_.size().height() / (double)qimage_.size().width();
	if(rect().width()*ratio <= rect().height())
		imgrect_.setRect(0, (rect().height()-rect().width()*ratio)/2, rect().width(), rect().width()*ratio);
	else
		imgrect_.setRect((rect().width()-rect().height()/ratio)/2, 0, rect().height()/ratio, rect().height());
	painter.drawImage(imgrect_, qimage_);

	painter.setPen(QPen(QBrush(QColor(0,0,0,180)),1,Qt::DashLine));
	painter.setBrush(QBrush(QColor(255,255,255,120)));
	painter.drawRect(selectionRect_);
}
